import os

from appdirs import user_data_dir

import click
import click_config_file

from storybro.cli.config import Config
from storybro.models.manager import ModelManager
from storybro.models.registry import ModelRegistry
from storybro.stories.manager import StoryManager
from .commands.models import models
from .commands.stories import stories
from .commands.play import play


data_dir = user_data_dir("storybro", "storybro")


@click.group()
@click.option('-r', '--model-registry',
              default="https://raw.githubusercontent.com/storybro/torrents/master/models.json")
@click.option('-m', '--models-path',
              show_default=os.path.join(data_dir, "models"),
              default=os.path.join(data_dir, "models"))
@click.option('-s', '--stories-path',
              show_default=os.path.join(data_dir, "stories"),
              default=os.path.join(data_dir, "stories"))
@click.option('-g', '--grammars-path',
              show_default=os.path.join(data_dir, "grammars"),
              default=os.path.join(data_dir, "grammars"))
@click.pass_context
@click_config_file.configuration_option()
def cli(ctx, model_registry, models_path, stories_path, grammars_path):
    Config.model_registry = ModelRegistry(model_registry)
    Config.models_path = models_path
    Config.grammars_path = grammars_path
    Config.stories_path = stories_path
    Config.story_manager = StoryManager(stories_path)
    Config.model_manager = ModelManager(models_path)
    ctx.obj = Config


def ep():
    cli.add_command(play)
    cli.add_command(models)
    cli.add_command(stories)
    cli()



import tensorflow as tf
from storybro.generation.gpt2 import model


def penalize_used(logits, output):

    # I want to change the indices of logits wherever the index is found in output
    change_tensor = tf.zeros_like(logits, dtype=logits.dtype)
    unique = tf.unique(output[0])[0]
    ones = tf.ones_like(unique, dtype=unique.dtype)
    indices = tf.expand_dims(unique, 1)

    updates = tf.scatter_nd(indices, ones, [logits.shape[1]])

    bool_tensor = tf.expand_dims(tf.cast(updates, tf.bool), 0)

    return tf.compat.v1.where(bool_tensor, logits * 0.85, logits)


def top_k_logits(logits, k):
    if k == 0:
        # no truncation
        return logits

    def _top_k():
        values, _ = tf.nn.top_k(logits, k=k)
        min_values = values[:, -1, tf.newaxis]
        return tf.where(
            logits < min_values,
            tf.ones_like(logits, dtype=logits.dtype) * -1e10,
            logits,
        )

    return tf.cond(tf.equal(k, 0), lambda: logits, lambda: _top_k(),)


def top_p_logits(logits, p):
    """Nucleus sampling"""
    batch, _ = logits.shape.as_list()
    sorted_logits = tf.sort(logits, direction="DESCENDING", axis=-1)
    cumulative_probs = tf.cumsum(tf.nn.softmax(sorted_logits, axis=-1), axis=-1)
    indices = tf.stack(
        [
            tf.range(0, batch),
            # number of indices to include
            tf.maximum(
                tf.reduce_sum(tf.cast(cumulative_probs <= p, tf.int32), axis=-1) - 1, 0
            ),
        ],
        axis=-1,
    )
    min_values = tf.gather_nd(sorted_logits, indices)
    return tf.where(logits < min_values, tf.ones_like(logits) * -1e10, logits,)


def sample_sequence(
    *,
    hparams,
    length,
    start_token=None,
    batch_size=None,
    context=None,
    temperature=1,
    top_k=0,
    top_p=1
):
    if start_token is None:
        assert context is not None, "Specify exactly one of start_token and context!"
    else:
        assert context is None, "Specify exactly one of start_token and context!"
        context = tf.fill([batch_size, 1], start_token)

    def step(hparams, tokens, past=None):
        lm_output = model.model(
            hparams=hparams, X=tokens, past=past, reuse=tf.AUTO_REUSE
        )

        logits = lm_output["logits"][:, :, : hparams.n_vocab]
        presents = lm_output["present"]
        presents.set_shape(model.past_shape(hparams=hparams, batch_size=batch_size))
        return {
            "logits": logits,
            "presents": presents,
        }

    with tf.name_scope("sample_sequence"):

        def body(past, prev, output):
            next_outputs = step(hparams, prev, past=past)
            logits = next_outputs["logits"][:, -1, :] / tf.to_float(temperature)
            logits = penalize_used(logits, output)
            logits = top_k_logits(logits, k=top_k)
            logits = top_p_logits(logits, p=top_p)
            samples = tf.multinomial(logits, num_samples=1, output_dtype=tf.int32)
            return [
                next_outputs["presents"]
                if past is None
                else tf.concat([past, next_outputs["presents"]], axis=-2),
                samples,
                tf.concat([output, samples], axis=1),
            ]

        past, prev, output = body(None, context, context)

        def cond(*args):
            return True

        _, _, tokens = tf.while_loop(
            cond=cond,
            body=body,
            maximum_iterations=length - 1,
            loop_vars=[past, prev, output],
            shape_invariants=[
                tf.TensorShape(
                    model.past_shape(hparams=hparams, batch_size=batch_size)
                ),
                tf.TensorShape([batch_size, None]),
                tf.TensorShape([batch_size, None]),
            ],
            back_prop=False,
        )

        return tokens



import os
import shutil

from storybro.utils import find_files, find_dirs


class Model:

    def __init__(self, root_path):
        self.root_path = root_path

    def filename_for(self, filename):
        return os.path.join(self.root_path, filename)

    def flatten(self):
        for filename in find_files(self.root_path):
            if os.path.dirname(filename) != self.root_path:
                shutil.move(filename, self.root_path)
        for dirname in find_dirs(self.root_path):
            shutil.rmtree(dirname)

    @property
    def name(self):
        return os.path.basename(self.root_path)



import os

from storybro.models.model import Model
from storybro.utils import find_child_dirs


class ModelManager:
    def __init__(self, root_path):
        self.root_path = root_path
        for model in self.models.values():
            model.flatten()

    @property
    def models(self):
        models = dict()

        for model_name in find_child_dirs(self.root_path):
            model_path = os.path.join(self.root_path, model_name)
            models[model_name] = Model(model_path)

        return models




import textwrap

from storybro.play.settings import PlayerSettings
from storybro.stories.block import Block
from storybro.stories.story import Story
from storybro.story.utils import cut_trailing_sentence


class BlockFormatter:
    def __init__(self, settings: PlayerSettings):
        self.settings: PlayerSettings = settings

    # before saving to disk

    def filter_block(self, block: Block) -> Block:
        if block.attrs.get('type') == 'input':
            return self.filter_input(block)
        else:
            return self.filter_output(block)

    def filter_input(self, block: Block) -> Block:
        return Block(block.text.strip(' \t\n\r'), attrs=block.attrs.copy())

    def filter_output(self, block: Block) -> Block:
        text = cut_trailing_sentence(block.text)

        if len(text) == 0:
            text = ""
        else:
            text = text.replace('."', '".')
            text = text.replace("#", "")
            text = text.replace("*", "")
            text = text.replace("\n\n", "\n")

            text = text.strip(' \t\n\r')

            if not text.endswith("."):
                text = text + "."

        return Block(text, attrs=block.attrs.copy())

    # before sending to generator

    def process_block(self, block: Block) -> Block:
        if block.attrs.get('type') == 'input':
            return self.process_input(block)
        else:
            return self.process_output(block)

    def process_input(self, block: Block) -> Block:
        return Block("\n> " + block.text + "\n", block.attrs.copy())

    def process_output(self, block: Block) -> Block:
        return Block(block.text, block.attrs.copy())

    def process_story(self, story: Story) -> str:
        return "".join([self.process_block(b).text for b in story.blocks])

    # before sending to user

    def render_block(self, block: Block) -> Block:
        if block.attrs.get('type') == 'input':
            return self.render_input(block)
        else:
            return self.render_output(block)

    def fill_text(self, text, is_input, is_pinned):
        input_icon = self.settings.icon_for_input if is_input else self.settings.icon_for_output
        pin_icon = self.settings.icon_for_pins if is_pinned else "  "
        initial_indent = input_icon + pin_icon + " "
        return textwrap.fill(
            text,
            width=self.settings.fill_width,
            replace_whitespace=True,
            drop_whitespace=True,
            initial_indent=initial_indent,
            subsequent_indent="     ",
            fix_sentence_endings=True)

    def render_input(self, block: Block) -> Block:
        rendered_input = block.text
        is_input = block.attrs.get('type') == 'input'
        is_pinned = block.attrs.get('pinned')

        if self.settings.fill_width:
            rendered_input = "\n".join([self.fill_text(l, is_input, is_pinned) for l in rendered_input.splitlines()])

        return Block(rendered_input, block.attrs.copy())

    def render_output(self, block: Block) -> Block:
        rendered_output = block.text
        is_input = block.attrs.get('type') == 'input'
        is_pinned = block.attrs.get('pinned')

        if self.settings.fill_width:
            rendered_output = "\n".join([self.fill_text(l, is_input, is_pinned) for l in rendered_output.splitlines()])

        rendered_output = f"{self.settings.top_separator}\n{rendered_output}\n{self.settings.bottom_separator}"
        return Block(rendered_output, block.attrs.copy())

    def render_story(self, story: Story, last_n: int = 0) -> str:
        blocks = story.blocks
        if last_n:
            blocks = blocks[-last_n:]
        return "\n".join([self.render_block(b).text for b in blocks])



import os


def find_files(path):
    for d, dirs, files in os.walk(path):
        for f in files:
            yield os.path.join(d, f)


def find_dirs(path):
    for d, dirs, files in os.walk(path):
        for path in dirs:
            yield os.path.join(d, path)


def find_child_dirs(path):
    try:
        return next(os.walk(path))[1]
    except StopIteration:
        return []



import os
import urllib.request
import json


class ModelRegistry:

    def __init__(self, url):
        self.url = url
        self.models = self.fetch(url)

    def fetch(self, url):
        response = urllib.request.urlopen(url)
        data = response.read()
        text = data.decode('utf-8')
        return json.loads(text)

    def refresh(self):
        self.models = self.fetch(self.url)



import os
import sys

import click

from storybro.play.block_formatter import BlockFormatter
from storybro.play.player import Player
from storybro.play.settings import PlayerSettings


@click.command()
@click.argument('story-name', required=True)
@click.option('-m', '--model-name', default="model_v5")
@click.option('--memory', type=int)
@click.option('--max-repeats', type=int)
@click.option('--icon-for-input')
@click.option('--top-separator')
@click.option('--bottom-separator')
@click.option('--fill-width', type=int)
@click.option('--force-cpu', '-f', is_flag=True, help="Force the model to run on the CPU")
@click.pass_obj
def play(config,
         story_name,
         model_name,
         memory, max_repeats,
         icon_for_input,
         top_separator, bottom_separator,
         fill_width,
         force_cpu):

    if force_cpu:
        os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

    model = config.model_manager.models.get(model_name)
    if not model:
        click.echo(f"Model `{model_name}` is not installed.")
        return

    for name in config.story_manager.stories:
        print(f"- {name}")

    story = config.story_manager.stories.get(story_name)
    if not story:
        if not click.confirm('Story does not exist. Create new story?'):
            return
        story = config.story_manager.new_story(story_name)

    sys.argv = ['storybro']
    settings = PlayerSettings(memory, max_repeats, icon_for_input, top_separator, bottom_separator, fill_width)
    formatter = BlockFormatter(settings)
    player = Player(model, story, settings, formatter)
    player.run()



import os
import shutil

import click


@click.command()
@click.argument('name')
@click.option('-t', '--torrent')
@click.option('--force', '-f', is_flag=True, help="Redownload if model already exists")
@click.pass_obj
def get(config, name, torrent, force):
    '''Download a model'''
    installed = config.model_manager.models

    if torrent is None:
        if name not in config.model_registry.models:
            click.echo(f"The model `{name}` wasn't found.")
            return

        torrent = config.model_registry.models[name]

    model = config.model_manager.models.get(name)

    if model:
        if not force:
            click.echo(f"The model `{name}` is already installed. Use -f to force re-download.")
            return
        else:
            shutil.rmtree(model.root_path)
            click.echo(f"Deleted local model `{name}`. Re-downloading.")

    click.echo(f"Downloading torrent: {torrent}")

    os.execlp('aria2c', 'aria2c',
        "--max-connection-per-server", "16",
        "--bt-max-peers", "500",
        "--seed-time", "0",
        "--summary-interval", "15",
        "--disable-ipv6",
        "--split", "64",
        "-d", config.models_path,
        torrent,
    )



import os
import shutil

from storybro.stories.story import Story
from storybro.utils import find_files


class DuplicateStoryError(Exception):
    def __init__(self, transcript_name):
        self.transcript_name = transcript_name


class StoryManager:
    def __init__(self, root_path):
        self.root_path = root_path
        os.makedirs(root_path, exist_ok=True)

    @property
    def stories(self):
        story_map = dict()

        for story_file in find_files(self.root_path):
            story_path = os.path.join(self.root_path, story_file)
            story_name = os.path.basename(story_file)
            story_map[story_name] = Story.load(story_path)

        return story_map

    def new_story(self, name):
        story = self.stories.get(name)

        if story:
            raise DuplicateStoryError(name)

        story_path = os.path.join(self.root_path, name)

        return Story(story_path)




class PlayerSettings:
    def __init__(self,
                 memory: int = None,
                 max_repeats: int = None,
                 icon_for_input: str = None,
                 icon_for_output: str = None,
                 icon_for_pins: str = None,
                 icon_for_delete: str = None,
                 top_separator: str = None,
                 bottom_separator: str = None,
                 fill_width: int = None
                 ):
        self.memory: int = memory or 20
        self.max_repeats: int = max_repeats or 5
        self.icon_for_input: str = icon_for_input or "📖"
        self.icon_for_output: str = icon_for_input or "💻"
        self.icon_for_pins: str = icon_for_pins or "📌"
        self.icon_for_delete: str = icon_for_delete or "❌"
        self.top_separator: str = ""
        self.bottom_separator: str = ""
        self.fill_width: int = fill_width or 80

    def settable(self):
        return {
            'debug': 'whether to print debug messages',
            'max_repeats': 'how many times the ai will attempt to regenerate',
            'icon_for_input': 'icon denoting user input blocks',
            'icon_for_output': 'icon denoting ai output blocks',
            'icon_for_pins': 'icon denoting pinned story blocks',
            'icon_for_delete': 'icon denoting story blocks to delete',
            'top_separator': 'string to use at the top of ai responses',
            'bottom_separator': 'string to use at the bottom of ai responses',
            'fill_width': 'number of characters to wrap ai responses. None to disable',
            'memory': 'how many blocks the ai will remember',
        }



import json
import os
import warnings

import numpy as np

import tensorflow as tf

from storybro.generation.gpt2 import model, encoder, sample
from storybro.story.utils import cut_trailing_sentence, remove_profanity

warnings.filterwarnings("ignore")

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


class GPT2Generator:
    def __init__(self, modelobj, generate_num=60, temperature=0.4, top_k=40, top_p=0.9, censor=True):
        self.generate_num = generate_num
        self.temp = temperature
        self.top_k = top_k

        self.top_p = top_p
        self.censor = censor

        self.model = modelobj

        self.batch_size = 1
        self.samples = 1

        self.enc = encoder.get_encoder(self.model.root_path)
        hparams = model.default_hparams()
        with open(os.path.join(self.model.root_path, "hparams.json")) as f:
            hparams.override_from_dict(json.load(f))
        seed = np.random.randint(0, 100000)

        config = tf.compat.v1.ConfigProto()
        config.gpu_options.allow_growth = True
        self.sess = tf.compat.v1.Session(config=config)

        self.context = tf.placeholder(tf.int32, [self.batch_size, None])
        # np.random.seed(seed)
        # tf.set_random_seed(seed)
        self.output = sample.sample_sequence(
            hparams=hparams,
            length=self.generate_num,
            context=self.context,
            batch_size=self.batch_size,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
        )

        saver = tf.train.Saver()
        ckpt = tf.train.latest_checkpoint(self.model.root_path)
        saver.restore(self.sess, ckpt)

    def prompt_replace(self, prompt):
        if len(prompt) > 0 and prompt[-1] == " ":
            prompt = prompt[:-1]

        return prompt

    def result_replace(self, result):
        result = cut_trailing_sentence(result)

        if len(result) == 0:
            return ""

        result = result.replace('."', '".')
        result = result.replace("#", "")
        result = result.replace("*", "")
        result = result.replace("\n\n", "\n")

        if self.censor:
            result = remove_profanity(result)

        return result

    def generate_raw(self, prompt):
        context_tokens = self.enc.encode(prompt)
        generated = 0
        for _ in range(self.samples // self.batch_size):
            out = self.sess.run(
                self.output,
                feed_dict={
                    self.context: [context_tokens for _ in range(self.batch_size)]
                },
            )[:, len(context_tokens) :]
            for i in range(self.batch_size):
                generated += 1
                text = self.enc.decode(out[i])
        return text

    def generate(self, prompt):

        debug_print = False
        filtered_prompt = self.prompt_replace(prompt)

        result = self.generate_raw(filtered_prompt)
        filtered_result = self.result_replace(result)

        if len(filtered_result) == 0:
            return self.generate(prompt)

        return filtered_result



import json
import os
import uuid

from storybro.story.utils import split_first_sentence, cut_trailing_sentence, console_print


class Story:
    def __init__(
        self, story_start, context="", seed=None, game_state=None, upload_story=False
    ):
        self.story_start = story_start
        self.context = context
        self.rating = -1
        self.upload_story = upload_story

        # list of actions. First action is the prompt length should always equal that of story blocks
        self.actions = []

        # list of story blocks first story block follows prompt and is intro story
        self.results = []

        # Only needed in constrained/cached version
        self.seed = seed
        self.choices = []
        self.possible_action_results = None
        self.uuid = None

        if game_state is None:
            game_state = dict()
        self.game_state = game_state
        self.memory = 20

    def __del__(self):
        if self.upload_story:
            self.save_to_storage()
            console_print("Game saved.")
            console_print(
                "To load the game, type 'load' and enter the following ID: " + self.uuid
            )

    def init_from_dict(self, story_dict):
        self.story_start = story_dict["story_start"]
        self.seed = story_dict["seed"]
        self.actions = story_dict["actions"]
        self.results = story_dict["results"]
        self.choices = story_dict["choices"]
        self.possible_action_results = story_dict["possible_action_results"]
        self.game_state = story_dict["game_state"]
        self.context = story_dict["context"]
        self.uuid = story_dict["uuid"]

        if "rating" in story_dict.keys():
            self.rating = story_dict["rating"]
        else:
            self.rating = -1

    def initialize_from_json(self, json_string):
        story_dict = json.loads(json_string)
        self.init_from_dict(story_dict)

    def add_to_story(self, action, story_block):
        self.actions.append(action)
        self.results.append(story_block)

    def latest_result(self):

        mem_ind = self.memory
        if len(self.results) < 2:
            latest_result = self.story_start
        else:
            latest_result = self.context
        while mem_ind > 0:

            if len(self.results) >= mem_ind:
                latest_result += self.actions[-mem_ind] + self.results[-mem_ind]

            mem_ind -= 1

        return latest_result

    def __str__(self):
        story_list = [self.story_start]
        for i in range(len(self.results)):
            story_list.append("\n" + self.actions[i] + "\n")
            story_list.append("\n" + self.results[i])

        return "".join(story_list)

    def to_json(self):
        story_dict = {}
        story_dict["story_start"] = self.story_start
        story_dict["seed"] = self.seed
        story_dict["actions"] = self.actions
        story_dict["results"] = self.results
        story_dict["choices"] = self.choices
        story_dict["possible_action_results"] = self.possible_action_results
        story_dict["game_state"] = self.game_state
        story_dict["context"] = self.context
        story_dict["uuid"] = self.uuid
        story_dict["rating"] = self.rating

        return json.dumps(story_dict)

    def save_to_storage(self):
        print("Saving to storage has been disabled due to abuse of the cloud bucket. Save will now be stored locally.")

        self.uuid = str(uuid.uuid1())

        save_path = "./saved_stories/"

        if not os.path.exists(save_path):
            os.makedirs(save_path)

        story_json = self.to_json()
        file_name = "story" + str(self.uuid) + ".json"
        f = open(os.path.join(save_path, file_name), "w")
        f.write(story_json)
        f.close()
        return self.uuid

    def load_from_storage(self, story_id):
        save_path = "./saved_stories/"

        if not os.path.exists(save_path):
            return "Error save not found."

        file_name = "story" + story_id + ".json"
        exists = os.path.isfile(os.path.join(save_path, file_name))
        if exists:
            with open(os.path.join(save_path, file_name), "r") as fp:
                game = json.load(fp)
                self.init_from_dict(game)
                return str(self)
        else:
            print("Save not found locally. Trying in the cloud bucket (only valid for saves before Dec 24 2019)")
            cmd = "gsutil cp gs://aidungeonstories/" + file_name + " " + save_path
            os.system(cmd)
            exists = os.path.isfile(os.path.join(save_path, file_name))


            if exists:
                with open(os.path.join(save_path, file_name), "r") as fp:
                    game = json.load(fp)
                self.init_from_dict(game)
                return str(self)
            else:
                return "Error save not found locally or in the cloud."

    def get_rating(self):
        while True:
            try:
                rating = input("Please rate the story quality from 1-10: ")
                rating_float = max(min(float(rating), 10), 1)
            except ValueError:
                print("Please return a valid number.")
            else:
                self.rating = rating_float
                return


class StoryManager:
    def __init__(self, generator):
        self.generator = generator
        self.story = None

    def start_new_story(
        self, story_prompt, context="", game_state=None, upload_story=False
    ):
        block = self.generator.generate(context + story_prompt)
        block = cut_trailing_sentence(block)
        self.story = Story(
            context + story_prompt + block,
            context=context,
            game_state=game_state,
            upload_story=upload_story,
        )
        return str(self.story)

    def load_new_story(self, story_id, upload_story=False):
        file_name = "story" + story_id + ".json"
        cmd = "gsutil cp gs://aidungeonstories/" + file_name + " ."
        os.system(cmd)
        exists = os.path.isfile(file_name)

        if not exists:
            return "Error save not found."

        with open(file_name, "r") as fp:
            game = json.load(fp)
        self.story = Story("", upload_story=upload_story)
        self.story.init_from_dict(game)
        return str(self.story)

    def load_story(self, story, from_json=False):
        if from_json:
            self.story = Story("")
            self.story.initialize_from_json(story)
        else:
            self.story = story
        return str(story)

    def json_story(self):
        return self.story.to_data()

    def story_context(self):
        return self.story.latest_result()


class UnconstrainedStoryManager(StoryManager):
    def act(self, action_choice):

        result = self.generate_result(action_choice)
        self.story.add_to_story(action_choice, result)
        return result

    def generate_result(self, action):
        block = self.generator.generate(self.story_context() + action)
        return block



import json
from typing import Set, List, Dict

from storybro.stories.block import Block
from storybro.story.utils import parse_slice


class Story:

    def __init__(self, path: str):
        self.path: str = path
        self.blocks: List[Block] = []
        self.attrs: Dict = {}

    def save(self):
        with open(self.path, 'w') as fobj:
            data = self.to_data()
            json.dump(data, fobj)

    @classmethod
    def load(cls, path):
        with open(path, 'r') as fobj:
            data = json.load(fobj)
        return cls.from_data(path, data)

    def to_data(self):
        return {
            'attrs': self.attrs,
            'blocks': [b.to_data() for b in self.blocks],
        }

    @classmethod
    def from_data(cls, path, data):
        story = cls(path)
        story.attrs = data["attrs"]
        story.blocks = [Block.from_data(d) for d in data["blocks"]]
        return story

    def filter_blocks(self, first_n=None, last_n=None, range=None):
        filtered = set([])

        if range and ":" in range:
            slice_obj: slice = parse_slice(range, bump_stop=True)
            filtered = filtered | set(self.blocks[slice_obj])

        if last_n:
            filtered = filtered | set(self.blocks[-last_n:])

        if first_n:
            filtered = filtered | set(self.blocks[:first_n])

        if not range and not last_n and not first_n:
            filtered = self.blocks

        return filtered


    def render_transcript(self, latest_n: int, specific_blocks: Set[int] = None):
        indices = specific_blocks or set([])

        total = len(self.blocks)

        if total == 0:
            return ""

        index = 0

        while index < total:
            if index in indices:
                index += 1
                continue

            if index > total - latest_n:
                indices.add(index)

            index += 1

        transcript = ""

        for index in sorted(indices):
            transcript += self.blocks[index]

        return transcript



import json
import os

from importlib_resources import read_text, path

import tracery
from tracery.modifiers import base_english


def apply_grammar(key, rules):
    grammar = tracery.Grammar(rules)
    grammar.add_modifiers(base_english)
    return grammar.flatten("#{}#".format(key))


def load_rules(grammar_file):
    with path('storybro.data.grammars', f"{grammar_file}.json") as grammar_path:
        with open(grammar_path, 'r') as fobj:
            return json.load(fobj)


def generate(grammar_file, character_type, key):
    """
    Provides a randomized prompt according to the rules in specified grammar file
    """
    rules = load_rules(grammar_file)
    artefact = apply_grammar("{}_{}".format(character_type, key), rules)
    return artefact


def direct(grammar_file, key):
    rules = load_rules(grammar_file)
    artefact = apply_grammar(key, rules)
    return artefact



from storybro.models.manager import ModelManager
from storybro.models.registry import ModelRegistry
from storybro.stories.manager import StoryManager


class Config:
    models_path: str = None
    stories_path: str = None
    grammars_path: str = None

    model_registry: ModelRegistry = None
    model_manager: ModelManager = None
    story_manager: StoryManager = None





import json


class Block:

    def __init__(self, text, attrs=None):
        self.text = text
        self.attrs = attrs or dict()

    def to_data(self) -> dict:
        return {
            'text': self.text,
            'attrs': self.attrs
        }

    @staticmethod
    def from_data(data: dict):
        return Block(data['text'], data['attrs'])



# coding: utf-8
import re
from difflib import SequenceMatcher

from importlib_resources import read_text

import yaml
from profanityfilter import ProfanityFilter

YAML_FILE = "story/story_data.yaml"


censored_words = read_text('storybro.data', 'censored_words.txt')
censored_words = [l.replace("\n", "") for l in censored_words.splitlines()]

pf = ProfanityFilter(custom_censor_list=censored_words)


def console_print(text, width=75):
    last_newline = 0
    i = 0
    while i < len(text):
        if text[i] == "\n":
            last_newline = 0
        elif last_newline > width and text[i] == " ":
            text = text[:i] + "\n" + text[i:]
            last_newline = 0
        else:
            last_newline += 1
        i += 1
    print(text)


def get_similarity(a, b):
    return SequenceMatcher(None, a, b).ratio()


def get_num_options(num):

    while True:
        choice = input("Enter the number of your choice: ")
        try:
            result = int(choice)
            if result >= 0 and result < num:
                return result
            else:
                print("Error invalid choice. ")
        except ValueError:
            print("Error invalid choice. ")


def player_died(text):
    """
    TODO: Add in more sophisticated NLP, maybe a custom classifier
    trained on hand-labelled data that classifies second-person
    statements as resulting in death or not.
    """
    lower_text = text.lower()
    you_dead_regexps = [
        "you('re| are) (dead|killed|slain|no more|nonexistent)",
        "you (die|pass away|perish|suffocate|drown|bleed out)",
        "you('ve| have) (died|perished|suffocated|drowned|been (killed|slain))",
        "you (\w* )?(yourself )?to death",
        "you (\w* )*(collapse|bleed out|chok(e|ed|ing)|drown|dissolve) (\w* )*and (die(|d)|pass away|cease to exist|(\w* )+killed)",
    ]
    return any(re.search(regexp, lower_text) for regexp in you_dead_regexps)


def player_won(text):
    lower_text = text.lower()
    won_phrases = [
        "you ((\w* )*and |)live happily ever after",
        "you ((\w* )*and |)live (forever|eternally|for eternity)",
        "you ((\w* )*and |)(are|become|turn into) ((a|now) )?(deity|god|immortal)",
        "you ((\w* )*and |)((go|get) (in)?to|arrive (at|in)) (heaven|paradise)",
        "you ((\w* )*and |)celebrate your (victory|triumph)",
        "you ((\w* )*and |)retire",
        "The rest is history...",
    ]
    return any(re.search(regexp, lower_text) for regexp in won_phrases)


def remove_profanity(text):
    return pf.censor(text)


def cut_trailing_quotes(text):
    num_quotes = text.count('"')
    if num_quotes % 2 is 0:
        return text
    else:
        final_ind = text.rfind('"')
        return text[:final_ind]


def split_first_sentence(text):
    first_period = text.find(".")
    first_exclamation = text.find("!")

    if first_exclamation < first_period and first_exclamation > 0:
        split_point = first_exclamation + 1
    elif first_period > 0:
        split_point = first_period + 1
    else:
        split_point = text[0:20]

    return text[0:split_point], text[split_point:]


def cut_trailing_action(text):
    lines = text.split("\n")
    last_line = lines[-1]
    if (
        "you ask" in last_line
        or "You ask" in last_line
        or "you say" in last_line
        or "You say" in last_line
    ) and len(lines) > 1:
        text = "\n".join(lines[0:-1])
    return text


def cut_trailing_sentence(text):
    text = standardize_punctuation(text)
    last_punc = max(text.rfind("."), text.rfind("!"), text.rfind("?"))
    if last_punc <= 0:
        last_punc = len(text) - 1

    et_token = text.find("<")
    if et_token > 0:
        last_punc = min(last_punc, et_token - 1)

    act_token = text.find(">")
    if act_token > 0:
        last_punc = min(last_punc, act_token - 1)

    text = text[:last_punc]

    text = cut_trailing_quotes(text)
    text = cut_trailing_action(text)
    return text


def replace_outside_quotes(text, current_word, repl_word):
    text = standardize_punctuation(text)

    reg_expr = re.compile(current_word + '(?=([^"]*"[^"]*")*[^"]*$)')

    output = reg_expr.sub(repl_word, text)
    return output


def is_first_person(text):

    count = 0
    for pair in first_to_second_mappings:
        variations = mapping_variation_pairs(pair)
        for variation in variations:
            reg_expr = re.compile(variation[0] + '(?=([^"]*"[^"]*")*[^"]*$)')
            matches = re.findall(reg_expr, text)
            count += len(matches)

    if count > 3:
        return True
    else:
        return False


def is_second_person(text):
    count = 0
    for pair in second_to_first_mappings:
        variations = mapping_variation_pairs(pair)
        for variation in variations:
            reg_expr = re.compile(variation[0] + '(?=([^"]*"[^"]*")*[^"]*$)')
            matches = re.findall(reg_expr, text)
            count += len(matches)

    if count > 3:
        return True
    else:
        return False


def capitalize(word):
    return word[0].upper() + word[1:]


def mapping_variation_pairs(mapping):
    mapping_list = []
    mapping_list.append((" " + mapping[0] + " ", " " + mapping[1] + " "))
    mapping_list.append(
        (" " + capitalize(mapping[0]) + " ", " " + capitalize(mapping[1]) + " ")
    )

    # Change you it's before a punctuation
    if mapping[0] is "you":
        mapping = ("you", "me")
    mapping_list.append((" " + mapping[0] + ",", " " + mapping[1] + ","))
    mapping_list.append((" " + mapping[0] + "\?", " " + mapping[1] + "\?"))
    mapping_list.append((" " + mapping[0] + "\!", " " + mapping[1] + "\!"))
    mapping_list.append((" " + mapping[0] + "\.", " " + mapping[1] + "."))

    return mapping_list


first_to_second_mappings = [
    ("I'm", "you're"),
    ("Im", "you're"),
    ("Ive", "you've"),
    ("I am", "you are"),
    ("was I", "were you"),
    ("am I", "are you"),
    ("wasn't I", "weren't you"),
    ("I", "you"),
    ("I'd", "you'd"),
    ("i", "you"),
    ("I've", "you've"),
    ("was I", "were you"),
    ("am I", "are you"),
    ("wasn't I", "weren't you"),
    ("I", "you"),
    ("I'd", "you'd"),
    ("i", "you"),
    ("I've", "you've"),
    ("I was", "you were"),
    ("my", "your"),
    ("we", "you"),
    ("we're", "you're"),
    ("mine", "yours"),
    ("me", "you"),
    ("us", "you"),
    ("our", "your"),
    ("I'll", "you'll"),
    ("myself", "yourself"),
]

second_to_first_mappings = [
    ("you're", "I'm"),
    ("your", "my"),
    ("you are", "I am"),
    ("you were", "I was"),
    ("are you", "am I"),
    ("you", "I"),
    ("you", "me"),
    ("you'll", "I'll"),
    ("yourself", "myself"),
    ("you've", "I've"),
]


def capitalize_helper(string):
    string_list = list(string)
    string_list[0] = string_list[0].upper()
    return "".join(string_list)


def capitalize_first_letters(text):
    first_letters_regex = re.compile(r"((?<=[\.\?!]\s)(\w+)|(^\w+))")

    def cap(match):
        return capitalize_helper(match.group())

    result = first_letters_regex.sub(cap, text)
    return result


def standardize_punctuation(text):
    text = text.replace("’", "'")
    text = text.replace("`", "'")
    text = text.replace("“", '"')
    text = text.replace("”", '"')
    return text


def first_to_second_person(text):
    text = " " + text
    text = standardize_punctuation(text)
    for pair in first_to_second_mappings:
        variations = mapping_variation_pairs(pair)
        for variation in variations:
            text = replace_outside_quotes(text, variation[0], variation[1])

    return capitalize_first_letters(text[1:])


def second_to_first_person(text):
    text = " " + text
    text = standardize_punctuation(text)
    for pair in second_to_first_mappings:
        variations = mapping_variation_pairs(pair)
        for variation in variations:
            text = replace_outside_quotes(text, variation[0], variation[1])

    return capitalize_first_letters(text[1:])


def parse_slice(value, bump_stop=False) -> slice:
    """
    Parses a `slice()` from string, like `start:stop:step`.
    """
    if value:
        parts = value.split(':')
        if len(parts) == 1:
            # slice(stop)
            parts = [None, parts[0]]
    else:
        parts = []

    if parts[1] and bump_stop:
        parts[1] = str(int(parts[1]) + 1)

    return slice(*[int(p) if p else None for p in parts])



from __future__ import unicode_literals

import argparse
import sys
import textwrap

import click
import cmd2
from cmd2 import with_argparser
from importlib_resources import read_text
from prompt_toolkit import PromptSession

from storybro.generation.gpt2.generator import GPT2Generator
from storybro.models.model import Model
from storybro.play.block_formatter import BlockFormatter
from storybro.play.settings import PlayerSettings
from storybro.stories.block import Block
from storybro.stories.story import Story
from storybro.story.utils import cut_trailing_sentence, parse_slice


class NoBlockCommitter(Exception):
    message = "No BlockCommitter set for Player."


class Player(cmd2.Cmd):
    """A simple cmd2 application."""

    def __init__(self, model: Model, story: Story, settings: PlayerSettings,
                 block_formatter: BlockFormatter):
        super().__init__()

        self.model: Model = model
        self.story: Story = story
        self.settings: PlayerSettings = settings
        self.block_formatter: BlockFormatter = block_formatter

        self.generator: GPT2Generator = None
        self.session: PromptSession = None

        self.remove_default_commands()
        self.setup_settables()
        self.allow_ansi = "Terminal"

    @property
    def prompt(self):
        return f"{self.settings.icon_for_input}   "

    def remove_default_commands(self):
        del cmd2.Cmd.do_alias
        del cmd2.Cmd.do_edit
        del cmd2.Cmd.do_history
        del cmd2.Cmd.do_macro
        del cmd2.Cmd.do_py
        del cmd2.Cmd.do_run_pyscript
        del cmd2.Cmd.do_run_script
        del cmd2.Cmd.do_shell
        del cmd2.Cmd.do_shortcuts

    def setup_generator(self, model):
        return GPT2Generator(model)

    def setup_settables(self):
        """Create pass-through properties on Player class for each setting in PlayerSettings"""
        self.settable = self.settings.settable()
        for attr in self.settable:
            setattr(Player, attr, property(lambda self, attr=attr: getattr(self.settings, attr),
                                           lambda self, value, attr=attr: setattr(self.settings, attr, value)))

    def display_splash(self):
        text = read_text('storybro.data', 'splash.txt')
        text += f"\n\n🕑 Initializing `{self.model.name}` model; Please wait..."
        self.poutput(text)

    def run(self):
        self.display_splash()
        self.generator = self.setup_generator(self.model)
        self.session = PromptSession()

        if self.story.blocks:
            self.poutput(self.block_formatter.render_story(self.story))

        sys.exit(self.cmdloop())

    def _input_line_to_statement(self, line: str) -> cmd2.Statement:
        line = line.strip()

        if line == "eof":
            line = "quit"
        elif not line.startswith("/"):
            line = f"commit {line}"
        else:
           line = line[1:]

        return super()._input_line_to_statement(line)

    def do_commit(self, input_text: str):
        print("🕑", end="\r")
        input_block = Block(input_text, attrs=dict(type='input'))
        input_block: Block = self.block_formatter.filter_block(input_block)
        if input_block.text:
            self.story.blocks.append(input_block)

        processed_story: str = self.block_formatter.process_story(self.story)
        output_text: str = self.generator.generate_raw(processed_story)

        output_block = Block(output_text, attrs=dict(type='output'))
        output_block = self.block_formatter.filter_block(output_block)
        self.story.blocks.append(output_block)

        print("  ", end="\r")

        rendered_block: Block = self.block_formatter.render_block(output_block)
        self.poutput(rendered_block.text)


    def do_save(self, text):
        self.story.save()
        self.poutput(f"Story saved: {self.story.path}.")

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-l', '--last-n', type=int, default=0, help='print the last [n] blocks')
    @with_argparser(argparser)
    def do_print(self, args):
        if self.story.blocks:
            self.poutput(self.block_formatter.render_story(self.story, args.last_n))

    def format_block_for_list(self, block, index, index_width):
        if block.attrs.get('type') == 'input':
            icon = self.settings.icon_for_input
        else:
            icon = self.settings.icon_for_output

        if block.attrs.get('pinned'):
            pin = self.settings.icon_for_pins
        else:
            pin = "  "

        index_label = str(index).rjust(index_width)

        text = block.text

        if self.settings.fill_width:
            text = textwrap.shorten(text, self.settings.fill_width, placeholder="")

        return f"{icon}{pin} {index_label}: {text}"

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-r', '--range', help="list the blocks in range", default=None, required=False)
    argparser.add_argument('-l', '--last_n', help="list the last n blocks", type=int, default=0, required=False)
    argparser.add_argument('-f', '--first_n', help="list the first n blocks", type=int, default=0, required=False)
    @with_argparser(argparser)
    def do_list(self, args):
        filtered = self.story.filter_blocks(args.first_n or None, args.last_n or None, args.range)
        n_blocks = len(self.story.blocks)
        index_width = len(f"{n_blocks}")

        for index, block in enumerate(self.story.blocks):
            if block in filtered:
                block_line = self.format_block_for_list(block, index, index_width)
                self.poutput(block_line)

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-r', '--range', help="list the blocks in range", default=None, required=False)
    argparser.add_argument('-l', '--last_n', help="list the last n blocks", type=int, default=0, required=False)
    argparser.add_argument('-f', '--first_n', help="list the first n blocks", type=int, default=0, required=False)
    @with_argparser(argparser)
    def do_delete(self, args):
        filtered = self.story.filter_blocks(args.first_n or None, args.last_n or None, args.range)
        n_blocks = len(self.story.blocks)
        index_width = len(f"{n_blocks}")

        for index, block in enumerate(self.story.blocks):
            prefix = "❌" if block in filtered else "  "
            block_line = self.format_block_for_list(block, index, index_width)
            self.poutput(f"{prefix}{block_line}")

        message = f"Delete these {len(filtered)} blocks?" if len(filtered) > 1 else "Delete this block?"
        if click.confirm(message):
            for block in filtered:
                self.story.blocks.remove(block)

            self.poutput("Done.")

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-r', '--range', help="list the blocks in range", default=None, required=False)
    argparser.add_argument('-l', '--last_n', help="list the last n blocks", type=int, default=0, required=False)
    argparser.add_argument('-f', '--first_n', help="list the first n blocks", type=int, default=0, required=False)
    @with_argparser(argparser)
    def do_pin(self, args):
        filtered = self.story.filter_blocks(args.first_n or None, args.last_n or None, args.range)
        n_blocks = len(self.story.blocks)
        index_width = len(f"{n_blocks}")

        for index, block in enumerate(self.story.blocks):
            if block.attrs.get('pinned'):
                continue

            prefix = self.settings.icon_for_pins if block in filtered else "  "
            block_line = self.format_block_for_list(block, index, index_width)
            self.poutput(f"{prefix}{block_line}")

        message = f"Pin these {len(filtered)} blocks?" if len(filtered) > 1 else "Pin this block?"
        if click.confirm(message):
            for block in filtered:
                block.attrs['pinned'] = True

            self.poutput("Done.")

    argparser = argparse.ArgumentParser()
    argparser.add_argument('-r', '--range', help="list the blocks in range", default=None, required=False)
    argparser.add_argument('-l', '--last_n', help="list the last n blocks", type=int, default=0, required=False)
    argparser.add_argument('-f', '--first_n', help="list the first n blocks", type=int, default=0, required=False)
    @with_argparser(argparser)
    def do_unpin(self, args):
        filtered = self.story.filter_blocks(args.first_n or None, args.last_n or None, args.range)
        n_blocks = len(self.story.blocks)
        index_width = len(f"{n_blocks}")

        for index, block in enumerate(self.story.blocks):
            if not block.attrs.get('pinned'):
                continue

            prefix = "❌" if block in filtered else "  "
            block_line = self.format_block_for_list(block, index, index_width)
            self.poutput(f"{prefix}{block_line}")

        message = f"Unpin these {len(filtered)} blocks?" if len(filtered) > 1 else "Unpin this block?"
        if click.confirm(message):
            for block in filtered:
                del block.attrs['pinned']

            self.poutput("Done.")

    argparser = argparse.ArgumentParser()
    @with_argparser(argparser)
    def do_stats(self, args):
        inputs = 0
        outputs = 0
        pinned = 0
        total_words = 0
        total_characters = 0

        for block in self.story.blocks:
            if block.attrs.get('type') == 'input':
                inputs += 1
            else:
                outputs += 1

            if block.attrs.get('pinned'):
                pinned += 1

            total_characters += len(block.text)
            total_words += len(block.text.split())

        self.poutput("Story stats:")
        self.poutput(f" - Inputs: {inputs}")
        self.poutput(f" - Outputs: {outputs}")
        self.poutput(f" - Pinned: {pinned}")
        self.poutput(f" - Words: {total_words}")
        self.poutput(f" - Characters: {total_characters}")




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['storybro',
 'storybro.cli',
 'storybro.cli.commands.models',
 'storybro.cli.commands.stories',
 'storybro.data',
 'storybro.data.grammars',
 'storybro.generation',
 'storybro.generation.gpt2',
 'storybro.models',
 'storybro.play',
 'storybro.stories',
 'storybro.story']

package_data = \
{'': ['*'], 'storybro.cli': ['commands/*']}

install_requires = \
['appdirs==1.4.3',
 'click-config-file==0.5.0',
 'click==7.0',
 'cmd2==0.9.22',
 'cryptography>=2.8,<2.9',
 'func_timeout==4.3.5',
 'google-cloud-storage==1.23.0',
 'gsutil==4.46',
 'importlib_resources>=1.0.2,<1.1.0',
 'numpy==1.18.0',
 'playsound==1.2.2',
 'profanityfilter==2.0.6',
 'prompt_toolkit==3.0.2',
 'pyyaml==5.2',
 'regex==2019.12.20',
 'tensorflow==1.15',
 'tracery==0.1.1']

entry_points = \
{'console_scripts': ['storybro = storybro:main']}

setup_kwargs = {
    'name': 'storybro',
    'version': '0.1.0',
    'description': 'A GPT-2 based AI assisted storytelling tool',
    'long_description': "# Storybro\n\nStorybro is a community maintained fork of [AI Dungeon 2](https://github.com/AIDungeon/AIDungeon). AI Dungeon was originally created by [Nick Walton](https://github.com/nickwalton).\n\nVisit our Wiki here: [Storybro wiki](https://github.com/storybro/storybro/wiki)\n\nRead more about AI Dungeon 2 and how it was built [here](https://pcc.cs.byu.edu/2019/11/21/ai-dungeon-2-creating-infinitely-generated-text-adventures-with-deep-learning-language-models/).\n\nThis fork aims to provide:\n- Improvements the original codebase\n- A command-line tool `storybro` for managing models, stories, etc\n- An improved interactive play mode\n- A model registry where models can be downloaded\n- An http web-service exposing most features\n- A Pip installable Python package\n\n### Note\n\nStorybro's AI can use your GPU or CPU. *A GPU will produce AI responses much faster than a CPU.* An nVidia GPU with 12 GB or more of memory, and CUDA installed, **is required** for GPU play. If you do not have such a GPU, you can play on your CPU. However, *each turn can take a couple of minutes or more* for the game to compose its response.\n\n# Playing\n\nThis README only covers installation. To learn how to play visit our [Wiki](http://github.com/storybro/storybro/wiki).\n\n# Installation\n\nInstalling Storybro requires the following software:\n\n- Python 3.4 - 3.7\n\nGrab the source code with Git and clone it to your machine:\n\n    git clone https://github.com/storybro/storybro/\n\n## Windows Installation\n\nStorybro comes with a few Windows Batch scripts to facilitate installation. If you'd like to install manually, see [Manual Installation](#manual-installation).\n\nStorybro uses [Chocolatey](http://chocolatey.org), a package manager, to install dependencies:\n\n1: Using an **ADMINISTRATOR** terminal from the root of this repo:\n\n\n    ./bin/windows/install/install-choco.bat\n\n2: Close your terminal and re-open it.\n    \n3: Now that Chocolatey is installed, install our dependencies and Storybro:\n\n\n    ./bin/windows/install/install-storybro.bat\n\n4: Once installation is done you should be in a shell. You can now use the `storybro` command:\n\n## Linux Installation\n\nStorybro comes with a few shell scripts to faciliate installation. If you'd like to install manually, see [Manual Installation](#manual-installation). \n\nTo install Storybro simply run the install script:\n\n    ./bin/linux/install/install\n    \nYou can now use [Poetry](https://python-poetry.org/) to enter a shell where you can use the `storybro` command:\n\n    poetry shell\n\n## Manual Installation\n\nStorybro is a Python application and uses [Poetry](https://python-poetry.org/) for its environment.\n\nInstall Poetry with Pip:\n\n    $ pip install poetry\n    \nUse Poetry to install Storybro:\n\n    $ poetry install\n    \nUse Poetry to run Storybro:\n\n    $ poetry run storybro\n    \nStorybro uses [Aria2](https://aria2.github.io/) to download models. Make sure that it is installed and on your `$PATH` if you intend to use Storybro to download models.\n\n\nCommunity\n------------------------\n\nStorybro is an open source project. Questions, discussion, and contributions are welcome. Contributions can be anything from new packages to bugfixes, documentation, or even new core features.\n\nResources:\n\n* **Reddit**: [r/AIDungeon](https://www.reddit.com/r/AIDungeon/)\n* **Discord**: [aidungeon discord](https://discord.gg/Dg8Vcz6)\n\n\nContributing\n------------------------\nContributing to Storybro is easy! Just send us a [pull request](https://help.github.com/articles/using-pull-requests/) from your fork. Make sure ``develop`` is the destination branch. \n\nStorybro uses a rough approximation of the [Git Flow](http://nvie.com/posts/a-successful-git-branching-model/) branching model.  The ``develop`` branch contains the latest contributions, and ``master`` is always tagged and points to the latest stable release.\n\nIf you're a contributor, make sure you're testing and playing on `develop`. That's where all the magic is happening (and where we hope bugs stop).\n",
    'author': 'Nick Walton',
    'author_email': 'unknown@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/storybro/storybro',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.8,<3.7.0',
}


setup(**setup_kwargs)



import click

from storybro.cli.config import Config


@click.command()
@click.pass_obj
def list(config: Config):
    '''List locally saved stories'''
    for key in config.story_manager.stories:
        print(key, end=" ")
        print("\n")



#!/usr/bin/env python3
import os
import random
import sys

import numpy as np
import yaml

from importlib_resources import read_text, path

from storybro.generation.gpt2.generator import GPT2Generator
from storybro.story import grammars
from storybro.story.story_manager import UnconstrainedStoryManager
from storybro.story.utils import console_print, get_num_options, player_died, player_won, get_similarity, \
    first_to_second_person, YAML_FILE

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"


def splash():
    print("0) New Game\n1) Load Game\n")
    choice = get_num_options(2)

    if choice == 1:
        return "load"
    else:
        return "new"


def random_story(story_data):
    # random setting
    settings = story_data["settings"].keys()
    n_settings = len(settings)
    n_settings = 2
    rand_n = random.randint(0, n_settings - 1)
    for i, setting in enumerate(settings):
        if i == rand_n:
            setting_key = setting

    # random character
    characters = story_data["settings"][setting_key]["characters"]
    n_characters = len(characters)
    rand_n = random.randint(0, n_characters - 1)
    for i, character in enumerate(characters):
        if i == rand_n:
            character_key = character

    # random name
    name = grammars.direct(setting_key, "character_name")

    return setting_key, character_key, name, None, None


def select_game():
    with path('storybro.data', 'story_data.yaml') as file_path:
        with open(file_path, "r") as file_obj:
            data = yaml.safe_load(file_obj)

    # Random story?
    print("Random story?")
    console_print("0) yes")
    console_print("1) no")
    choice = get_num_options(2)

    if choice == 0:
        return random_story(data)

    # User-selected story...
    print("\n\nPick a setting.")
    settings = data["settings"].keys()
    for i, setting in enumerate(settings):
        print_str = str(i) + ") " + setting
        if setting == "fantasy":
            print_str += " (recommended)"

        console_print(print_str)
    console_print(str(len(settings)) + ") custom")
    choice = get_num_options(len(settings) + 1)

    if choice == len(settings):
        return "custom", None, None, None, None

    setting_key = list(settings)[choice]

    print("\nPick a character")
    characters = data["settings"][setting_key]["characters"]
    for i, character in enumerate(characters):
        console_print(str(i) + ") " + character)
    character_key = list(characters)[get_num_options(len(characters))]

    name = input("\nWhat is your name? ")
    setting_description = data["settings"][setting_key]["description"]
    character = data["settings"][setting_key]["characters"][character_key]

    return setting_key, character_key, name, character, setting_description


def get_custom_prompt():
    context = ""
    console_print(
        "\nEnter a prompt that describes who you are and the first couple sentences of where you start "
        "out ex:\n 'You are a knight in the kingdom of Larion. You are hunting the evil dragon who has been "
        + "terrorizing the kingdom. You enter the forest searching for the dragon and see' "
    )
    prompt = input("Starting Prompt: ")
    return context, prompt


def get_curated_exposition(
    setting_key, character_key, name, character, setting_description
):
    name_token = "<NAME>"
    try:
        context = grammars.generate(setting_key, character_key, "context") + "\n\n"
        context = context.replace(name_token, name)
        prompt = grammars.generate(setting_key, character_key, "prompt")
        prompt = prompt.replace(name_token, name)
    except:
        context = (
            "You are "
            + name
            + ", a "
            + character_key
            + " "
            + setting_description
            + "You have a "
            + character["item1"]
            + " and a "
            + character["item2"]
            + ". "
        )
        prompt_num = np.random.randint(0, len(character["prompts"]))
        prompt = character["prompts"][prompt_num]

    return context, prompt


def instructions():
    text = "\nStorybro Instructions:"
    text += '\n Enter actions starting with a verb ex. "go to the tavern" or "attack the orc."'
    text += '\n To speak enter \'say "(thing you want to say)"\' or just "(thing you want to say)" '
    text += "\n\nThe following commands can be entered for any action: "
    text += '\n  "/revert"   Reverts the last action allowing you to pick a different action.'
    text += '\n  "/quit"     Quits the game and saves'
    text += '\n  "/reset"    Starts a new game and saves your current one'
    text += '\n  "/restart"  Starts the game from beginning with same settings'
    text += '\n  "/save"     Makes a new save of your game and gives you the save ID'
    text += '\n  "/load"     Asks for a save ID and loads the game if the ID is valid'
    text += '\n  "/print"    Prints a transcript of your adventure (without extra newline formatting)'
    text += '\n  "/help"     Prints these instructions again'
    text += '\n  "/censor off/on" to turn censoring off or on.'
    return text


def play_aidungeon_2(model_dir):

    upload_story = False

    print("\nInitializing Storybro! (This might take a few minutes)\n")
    generator = GPT2Generator(model_dir)
    story_manager = UnconstrainedStoryManager(generator)
    print("\n")

    starter = read_text('storybro.data', 'splash.txt')
    print(starter)

    while True:
        if story_manager.story != None:
            story_manager.story = None

        while story_manager.story is None:
            print("\n\n")
            splash_choice = splash()

            if splash_choice == "new":
                print("\n\n")
                (
                    setting_key,
                    character_key,
                    name,
                    character,
                    setting_description,
                ) = select_game()

                if setting_key == "custom":
                    context, prompt = get_custom_prompt()

                else:
                    context, prompt = get_curated_exposition(
                        setting_key, character_key, name, character, setting_description
                    )

                console_print(instructions())
                print("\nGenerating story...")

                result = story_manager.start_new_story(
                    prompt, context=context, upload_story=upload_story
                )
                print("\n")
                console_print(result)

            else:
                load_ID = input("What is the ID of the saved game? ")
                result = story_manager.load_new_story(
                    load_ID, upload_story=upload_story
                )
                print("\nLoading Game...\n")
                console_print(result)

        while True:
            sys.stdin.flush()
            action = input("> ").strip()
            if len(action) > 0 and action[0] == "/":
                split = action[1:].split(" ")  # removes preceding slash
                command = split[0].lower()
                args = split[1:]
                if command == "reset":
                    story_manager.story.get_rating()
                    break

                elif command == "restart":
                    story_manager.story.actions = []
                    story_manager.story.results = []
                    console_print("Game restarted.")
                    console_print(story_manager.story.story_start)
                    continue

                elif command == "quit":
                    story_manager.story.get_rating()
                    exit()

                elif command == "nosaving":
                    upload_story = False
                    story_manager.story.upload_story = False
                    console_print("Saving turned off.")

                elif command == "help":
                    console_print(instructions())

                elif command == "censor":
                    if len(args) == 0:
                        if generator.censor:
                            console_print("Censor is enabled.")
                        else:
                            console_print("Censor is disabled.")
                    elif args[0] == "off":
                        if not generator.censor:
                            console_print("Censor is already disabled.")
                        else:
                            generator.censor = False
                            console_print("Censor is now disabled.")

                    elif args[0] == "on":
                        if generator.censor:
                            console_print("Censor is already enabled.")
                        else:
                            generator.censor = True
                            console_print("Censor is now enabled.")

                    else:
                        console_print("Invalid argument: {}".format(args[0]))

                elif command == "save":
                    if upload_story:
                        id = story_manager.story.save_to_storage()
                        console_print("Game saved.")
                        console_print(
                            "To load the game, type 'load' and enter the "
                            "following ID: {}".format(id)
                        )
                    else:
                        console_print("Saving has been turned off. Cannot save.")

                elif command == "load":
                    if len(args) == 0:
                        load_ID = input("What is the ID of the saved game?")
                    else:
                        load_ID = args[0]
                    result = story_manager.story.load_from_storage(load_ID)
                    console_print("\nLoading Game...\n")
                    console_print(result)

                elif command == "print":
                    print("\nPRINTING\n")
                    print(str(story_manager.story))

                elif command == "revert":
                    if len(story_manager.story.actions) == 0:
                        console_print("You can't go back any farther. ")
                        continue

                    story_manager.story.actions = story_manager.story.actions[:-1]
                    story_manager.story.results = story_manager.story.results[:-1]
                    console_print("Last action reverted. ")
                    if len(story_manager.story.results) > 0:
                        console_print(story_manager.story.results[-1])
                    else:
                        console_print(story_manager.story.story_start)
                    continue

                else:
                    console_print("Unknown command: {}".format(command))

            else:
                if action == "":
                    action = ""
                    result = story_manager.act(action)
                    console_print(result)

                elif action[0] == '"':
                    action = "You say " + action

                else:
                    action = action.strip()

                    if "you" not in action[:6].lower() and "I" not in action[:6]:
                        action = action[0].lower() + action[1:]
                        action = "You " + action

                    if action[-1] not in [".", "?", "!"]:
                        action = action + "."

                    action = first_to_second_person(action)

                    action = "\n> " + action + "\n"

                result = "\n" + story_manager.act(action)
                if len(story_manager.story.results) >= 2:
                    similarity = get_similarity(
                        story_manager.story.results[-1], story_manager.story.results[-2]
                    )
                    if similarity > 0.9:
                        story_manager.story.actions = story_manager.story.actions[:-1]
                        story_manager.story.results = story_manager.story.results[:-1]
                        console_print(
                            "Woops that action caused the model to start looping. Try a different action to prevent that."
                        )
                        continue

                if player_won(result):
                    console_print(result + "\n CONGRATS YOU WIN")
                    story_manager.story.get_rating()
                    break
                elif player_died(result):
                    console_print(result)
                    console_print("YOU DIED. GAME OVER")
                    console_print("\nOptions:")
                    console_print("0) Start a new game")
                    console_print(
                        "1) \"I'm not dead yet!\" (If you didn't actually die) "
                    )
                    console_print("Which do you choose? ")
                    choice = get_num_options(2)
                    if choice == 0:
                        story_manager.story.get_rating()
                        break
                    else:
                        console_print("Sorry about that...where were we?")
                        console_print(result)

                else:
                    console_print(result)



import click


@click.command()
@click.pass_obj
def list(config):
    '''List models available for download'''
    installed = config.model_manager.models.keys()
    available = config.model_registry.models.keys()
    keys = set(installed).union(set(available))
    for key in keys:
        bullet = "-" if key not in installed else "*"
        print(f"{bullet} {key}")
